! Program created by Gustavo Catao Alves at
! the Stanford Exploration Project
! Jan, 13th, 2015

! This is a reformulation of the original solution, but using the
! Velocity-stress method instead of the velocity-stress

! The grid positions are changed in this staggered-grid approach:
!  (i,j)                       (i+1/2,j)
!   sigmaxz,mi===================Vz,rho
!    |                             |
!    |                             |
!    |                             |
!    |                             |
!    |                             |
!    |                             |
!    |                             |
!    |                             |
!   Vx,rho=======sigmaxx,sigmazz,(lambda+2mi),lambda
!  (i,j+1/2)                 (i+1/2,j+1/2)


! FD elastic operator module
! This is the main subroutine that calls
! the FD derivatives, sources, etc.
module elastic_mod

  use source_mod
  use seis_mod
  use elasticlib_mod
  use snap_mod
  !use pml_mod
  use omp_lib
  !use clib_mod

  !use c_lib

  implicit none
  integer, private :: nt, nx, nz, abc, zrec, ix
  integer, private :: xsource, zsource, nsource, nsnap, nseis
  real, private    :: dt, dx, dz, fp
  !real*16, private :: source
  real, private :: source
  logical, private :: free_surface

  contains

  ! Initialization of all variables needed by this subroutine
  subroutine elastic_init(nz_in,dz_in,nx_in,dx_in,nt_in,dt_in,xsource_in,  &
                          zsource_in,nsource_in, fp_in, nsnap_in, nseis_in,&
                          abc_in, zrec_in, free_surface_in)

    integer             :: nz_in,nx_in,nt_in,xsource_in,zsource_in,nsource_in
    integer             :: nsnap_in, nseis_in, abc_in, zrec_in
    real                :: dz_in,dx_in,dt_in, fp_in
    logical             :: free_surface_in

    nz = nz_in
    dz = dz_in
    nt = nt_in
    dt = dt_in
    nx = nx_in
    dx = dx_in
    fp = fp_in
    xsource = xsource_in
    zsource = zsource_in
    nsource = nsource_in
    nsnap   = nsnap_in
    nseis = nseis_in
    abc = abc_in
    zrec = zrec_in
    free_surface = free_surface_in

  end subroutine

  ! Function that is called by Main
  function elastic_op(vp, vs, rho, seis_Vx, seis_Vz, seis_P, snap_Vx, snap_Vz, snap_P, lame) result(stat)
    real,dimension(:,:)         :: vp, vs, rho
    real,dimension(:,:)         :: seis_Vx, seis_Vz, seis_P
    real,dimension(:,:,:)       :: snap_Vx, snap_Vz, snap_P
    integer                     :: stat
    logical,intent(in),optional :: lame

    stat=1
    call elastic_op2(vp, vs, rho, seis_Vx, seis_Vz, seis_P, snap_Vx, snap_Vz, snap_P, lame)
    stat=0
  end function

  ! Subroutine used by function
  subroutine elastic_op2(vp, vs, rho, seis_Vx, seis_Vz, seis_P, snap_Vx, snap_Vz, snap_P,lame)
    integer                             :: it, ix, iz, dsnap, dseis
    integer                             :: countsnap,countseis, temp
    real, dimension(nz,nx)              :: vp,vs,rho
    real, dimension(nseis,nx-2*abc)     :: seis_Vx, seis_Vz, seis_P
    real, dimension(nz,nx,nsnap)        :: snap_Vx,snap_Vz, snap_P

    real, allocatable, dimension(:,:)   :: mi, l, l2mi
    !real*16, allocatable, dimension(:,:):: Vx,Vz,sigmaxx,sigmazz,sigmaxz
    !real*16, allocatable, dimension(:,:):: dxsigma,dzsigma,exx, ezz, exz1,exz2
    real, allocatable, dimension(:,:)   :: Vxold, Vzold
    real, allocatable, dimension(:,:)   :: sigmaxxold,sigmazzold,sigmaxzold
    real, allocatable, dimension(:,:)   :: Vxcur, Vzcur
    real, allocatable, dimension(:,:)   :: sigmaxxcur,sigmazzcur,sigmaxzcur
    real, allocatable, dimension(:,:)   :: Vxnew, Vznew
    real, allocatable, dimension(:,:)   :: sigmaxxnew,sigmazznew,sigmaxznew
    real, allocatable, dimension(:,:)   :: exx, ezz, exz1,exz2
    real, allocatable, dimension(:,:)   :: rho_Vx, rho_Vz
    real, allocatable, dimension(:,:)   :: temp1,temp2,temp3
    real                                :: aux0,aux1, t, nsource2

    integer                             :: stat
    logical,intent(in),optional         :: lame

    !Variables to measure runtime
    real*8                                :: start_time,end_time

    !Memory allocation

    !Allocate lame parameters
    allocate(mi(nz,nx))
    allocate(l(nz,nx))
    allocate(l2mi(nz,nx))
    allocate(rho_Vx(nz,nx))
    allocate(rho_Vz(nz,nx))

    !Allocate Velocities and stresses
    allocate(Vxold(nz,nx))
    allocate(Vzold(nz,nx))
    allocate(sigmaxxold(nz,nx))
    allocate(sigmazzold(nz,nx))
    allocate(sigmaxzold(nz,nx))
    allocate(Vxcur(nz,nx))
    allocate(Vzcur(nz,nx))
    allocate(sigmaxxcur(nz,nx))
    allocate(sigmazzcur(nz,nx))
    allocate(sigmaxzcur(nz,nx))
    allocate(Vxnew(nz,nx))
    allocate(Vznew(nz,nx))
    allocate(sigmaxxnew(nz,nx))
    allocate(sigmazznew(nz,nx))
    allocate(sigmaxznew(nz,nx))

    allocate(temp1(nz,nx))
    allocate(temp2(nz,nx))
    allocate(temp3(nz,nx))

    !Allocate strains
    allocate(exx(nz,nx))
    allocate(ezz(nz,nx))
    allocate(exz1(nz,nx))
    allocate(exz2(nz,nx))

    !Initialize values
    mi         =0.0
    l          =0.0
    l2mi       =0.0
    Vxold         =0.0
    Vzold         =0.0
    sigmaxxold    =0.0
    sigmazzold    =0.0
    sigmaxzold    =0.0
    Vxcur         =0.0
    Vzcur         =0.0
    sigmaxxcur    =0.0
    sigmazzcur    =0.0
    sigmaxzcur    =0.0
    Vxnew         =0.0
    Vznew         =0.0
    sigmaxxnew    =0.0
    sigmazznew    =0.0
    sigmaxznew    =0.0
    rho_Vx     =0.0
    rho_Vz     =0.0
    exx        =0.0
    ezz        =0.0
    exz1       =0.0
    exz2       =0.0
    dseis      =(nt/nseis)+1
    dsnap      =nt/nsnap
    countsnap  =1
    countseis  =1
    nsource2   =2*nint(sqrt(6.0)/3.14159265/fp/dt)

    temp1      =0.0
    temp2      =0.0
    temp3      =0.0

    !Define Lame parameters
    !These model properties are still in the original grid, centered at i,j
    do ix=1,nx
      do iz=1,nz
        if(lame)then
          l(iz,ix)=vp(iz,ix)
          mi(iz,ix)=vs(iz,ix)
        else
          mi(iz,ix)=vs(iz,ix)*vs(iz,ix)*rho(iz,ix)
          !mi(iz,ix)=0.
          l(iz,ix) =vp(iz,ix)*vp(iz,ix)*rho(iz,ix)-2.0*mi(iz,ix)
          l2mi(iz,ix)=l(iz,ix)+2.0*mi(iz,ix)
        endif
      enddo
    enddo

    !Calculate inverse of density
    !These also calculate the values of density at the velocity staggered grid
    !Rho_Vx

    aux1=(dt/dx)

    !Rho_vx
    do ix=1,nx
      do iz=1,nz-1
        rho_Vx(iz,ix)=aux1*1.0/(0.5*(rho(iz+1,ix) + rho(iz,ix)))
      enddo
    enddo
    !Rho_vz
    do ix=1,nx-1
      do iz=1,nz
        rho_Vz(iz,ix)=aux1*1.0/(0.5*(rho(iz,ix+1) + rho(iz,ix)))
      enddo
    enddo
    !Right side
    do iz=1,nz
      rho_Vz(iz,nx)=aux1*(1.0/rho(iz,nx))
    enddo
    !Bottom
    do ix=1,nx
      rho_Vx(nz,ix)=aux1*(1.0/rho(nz,ix))
    enddo
    
    !Calculate weigthed average for the lambda variable
    !These put lambda and (lambda+2mi) in their respective staggered grids
    do ix=2,nx-1
      do iz=2,nz-1
        l(iz,ix)   =aux1*4.0/((1.0/l(iz,ix))+(1.0/l(iz+1,ix))+(1.0/l(iz,ix+1))+(1.0/l(iz+1,ix+1)))
        if (mi(iz,ix)==0.or.mi(iz+1,ix)==0.or.mi(iz,ix+1)==0.or.mi(iz+1,ix+1)==0) then
          mi(iz,ix) = 0.0
        else
          mi(iz,ix)  =aux1*4.0/((1.0/mi(iz,ix))+(1.0/mi(iz+1,ix))+(1.0/mi(iz,ix+1))+(1.0/mi(iz+1,ix+1)))
        endif
        l2mi(iz,ix)=l(iz,ix)+2.0*mi(iz,ix)
      enddo
    enddo


    ! TEST MODEL PROPERTIES VALUES
    write(0,*)"============================================================="
    write(0,*)"aux1= dt/dx",aux1
    write(0,*)"rho(xsource,zsource)=",rho(xsource,zsource)
    write(0,*)"rho_Vx(xsource,zsource)=",rho_Vx(xsource,zsource)
    write(0,*)"rho_Vz(xsource,zsource)=",rho_Vz(xsource,zsource)
    write(0,*)"vp(xsource,zsource)=",vp(xsource,zsource)
    write(0,*)"vs(xsource,zsource)=",vs(xsource,zsource)
    write(0,*)"l(xsource,zsource)=",l(xsource,zsource)
    write(0,*)"mi(xsource,zsource)=",mi(xsource,zsource)
    write(0,*)"l2mi(xsource,zsource)=",l2mi(xsource,zsource)
    write(0,*)"Free surface=",free_surface
    write(0,*)"============================================================="

    !Initialize subroutines
    call source_init(fp,dt,dx,nsource,nz,nx)
    call seis_init(nz,nx, abc, zrec, seis_Vx, seis_Vz, seis_P)
    call elasticmod_init(nz,nx,rho_Vx,rho_Vz,l,mi,l2mi)
    !call pml_init(nz,nx,dx,dt,abc,vp,free_surface)
    call snap_init(nz,nx,abc,snap_Vx,snap_Vz, snap_P)
    
    !Time loop for 2nd order time
    write(0,*)"============================================================="
    write(0,*)"BEGIN OF TIME MODELING"
    write(0,*)"============================================================="
    write(0,*)"Processing:"
    write(0,*)"0%           25%            50%           75%            100%"
    write(0,'(a)',advance="no")"["

    start_time = omp_get_wtime()

    do it=1,nt
      if (it<nsource) then
        call source_op2(it, source)
      else
        source=0.0
      endif

      source = 0.7E7*source*(dt/(dx*dx))*0.25
      !Vx
      call Ax(.false.,.false.,sigmaxxcur,Vxnew)
      call Bx(.false.,.true.,sigmaxzcur,Vxnew)
      Vxnew = Vxnew + Vxold
      !call explosive_source(.false.,.false.,source,xsource,zsource,temp1)
      !call C(.false.,.true.,temp1,Vxnew)
      !Vz
      call Bz(.false.,.false.,sigmazzcur,Vznew)
      call Az(.false.,.true.,sigmaxzcur,Vznew)
      Vznew = Vznew + Vzold
      !call explosive_source(.false.,.false.,source,xsource,zsource,temp1)
      !call C(.false.,.true.,temp1,Vznew)
      !sigmaxx
      call D(.false.,.false.,Vxcur,sigmaxxnew)
      call E(.false.,.true.,Vzcur,sigmaxxnew)
      sigmaxxnew = sigmaxxnew + sigmaxxold
      call explosive_source(.false.,.true.,source,xsource,zsource,sigmaxxnew)
      !sigmazz
      call F(.false.,.false.,Vxcur,sigmazznew)
      call G(.false.,.true.,Vzcur,sigmazznew)
      sigmazznew = sigmazznew + sigmazzold
      call explosive_source(.false.,.true.,source,xsource,zsource,sigmazznew)
      !sigmaxz
      call H(.false.,.false.,Vxcur,sigmaxznew)
      call J(.false.,.true.,Vzcur,sigmaxznew)
      sigmaxznew = sigmaxznew + sigmaxzold
      !call explosive_source(.false.,.true.,source,xsource,zsource,sigmaxznew)

      !Write seismograms
      if (it.gt.nsource2.and.mod(it,dseis)==0) then
        call seis_op2(countseis, Vxnew, Vznew)
        call seis_op3(countseis, sigmaxxnew, sigmazznew)
        countseis = countseis + 1
        !write(0,*)"Seismogram time sample"
      endif

      !Write snapshots
      if (mod(it,dsnap)==0.and.countsnap.le.nsnap) then
        call snap_op2(countsnap,Vxnew,Vznew)
        call snap_op3(countsnap,sigmaxxnew,sigmazznew)
        countsnap = countsnap + 1
      endif

      Vxold = Vxcur
      Vzold = Vzcur
      sigmaxxold = sigmaxxcur
      sigmazzold = sigmazzcur
      sigmaxzold = sigmaxzcur

      Vxcur = Vxnew
      Vzcur = Vznew
      sigmaxxcur = sigmaxxnew
      sigmazzcur = sigmazznew
      sigmaxzcur = sigmaxznew

      !Write to terminal
      if (mod(it,(nt/59))==0) then
        write(0,'(a)',advance="no")"#"
        close(0)
      endif
    enddo
    write(0,'(a)',advance="yes")"]"
    write(0,*)"Last seismogram sample is:",countseis

    end_time = omp_get_wtime()
  ! Write outputs to the terminal
  write(0,*)"============================================================="
  write(0,*)"TOTAL ELAPSED TIME"
  write(0,*)"============================================================="
  write(0,*)"Total=",end_time - start_time, "seconds."
  write(0,*)"============================================================="
  end subroutine

end module
